@@ -26,7 +26,7 @@
 #include <sys/time.h> /* TODO: replace timeofday with g_timers */
 #include <gdk/gdkwindow.h>
 #include <gtk/gtk.h>
-#ifndef USE_GDKPIXBUF_RENDER_TO_DRAWABLE
+#ifdef USE_GDKPIXBUF_RENDER_TO_DRAWABLE
 #include <gdk/gdkpixbuf.h>
 #else
 #include <gdk/gdk.h>
@@ -217,6 +217,10 @@ static void zoom_region_paint_pixmap (ZoomRegion *zoom_region, GdkRectangle *rec
 static int  zoom_region_update_pointer_timeout (gpointer data);
 static void zoom_region_recompute_exposed_viewport (ZoomRegion *zoom_region);
 
+static GdkRectangle zoom_region_rect_from_bounds (ZoomRegion *zoom_region,
+						  const GNOME_Magnifier_RectBounds *bounds);
+static ZoomRegionPixmapCreationError zoom_region_create_pixmap (ZoomRegion *zoom_region);
+
 void
 reset_timing_stats()
 {
@@ -597,8 +601,12 @@ static GdkRectangle
 _rectangle_clip_to_rectangle (GdkRectangle area,
 			      GdkRectangle clip_rect)
 {
-	/* TODO: Implement this! */
-	return area;
+        GdkRectangle clipped;
+        clipped.x = MAX (area.x, clip_rect.x);
+        clipped.y = MAX (area.y, clip_rect.y);
+        clipped.width = MIN ((area.x + area.width), (clip_rect.x + clip_rect.width)) - clipped.x;
+        clipped.height = MIN ((area.y + area.height), (clip_rect.y + clip_rect.height)) - clipped.y;
+	return clipped;
 }
 
 static GdkRectangle
@@ -622,6 +630,7 @@ zoom_region_clip_to_source (ZoomRegion *zoom_region,
     if (zoom_region && zoom_region->priv && zoom_region->priv->parent)
     {
 	source_rect_ptr = &((Magnifier *)zoom_region->priv->parent)->source_bounds;
+        DEBUG_RECT ("clipping to source bounds", zoom_region_rect_from_bounds (zoom_region, source_rect_ptr)); 
 	return _rectangle_clip_to_bounds (area, source_rect_ptr);
     }
     return area;
@@ -660,11 +669,18 @@ zoom_region_clip_to_viewport (ZoomRegion *zoom_region,
 }
 #endif
 
-static const GdkRectangle
-zoom_region_clip_to_target (ZoomRegion *zoom_region,
-			    GdkRectangle area)
+static GdkRectangle
+zoom_region_clip_to_scaled_pixmap (ZoomRegion *zoom_region,
+				   GdkRectangle area)
 {
-	return _rectangle_clip_to_bounds (area, &zoom_region->priv->source_area);
+        GdkRectangle pixmap_area = {0, 0, 0, 0};
+	if (zoom_region->priv && zoom_region->priv->pixmap)
+	{
+	    gdk_drawable_get_size (zoom_region->priv->pixmap, &pixmap_area.width, &pixmap_area.height);
+	    return _rectangle_clip_to_rectangle (area, pixmap_area);
+	}
+	else
+	    return area;
 }
 
 static GdkRectangle
@@ -760,10 +776,12 @@ zoom_region_queue_update (ZoomRegion *zoom_region,
 
 	zoom_region->priv->q =
 		g_list_prepend (zoom_region->priv->q, rect);
-	g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
-			 zoom_region_process_updates,
-			 zoom_region,
-			 NULL);
+	if (zoom_region->priv && zoom_region->priv->update_handler_id == 0)
+		zoom_region->priv->update_handler_id = 
+			g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
+					 zoom_region_process_updates,
+					 zoom_region,
+					 NULL);
 }
 
 static void
@@ -772,11 +790,17 @@ zoom_region_update_current (ZoomRegion *zoom_region)
 #ifdef ZOOM_REGION_DEBUG
 	g_assert (zoom_region->alive);
 #endif
-	if (zoom_region->priv && GDK_IS_DRAWABLE (zoom_region->priv->pixmap))
-	    zoom_region_update (zoom_region,
-			    zoom_region_source_rect_from_view_bounds (
-				    zoom_region,
-				    &zoom_region->priv->exposed_viewport));
+	if (zoom_region->priv)
+	{
+		gboolean pixmap_valid = GDK_IS_DRAWABLE (zoom_region->priv->pixmap);
+		if (!pixmap_valid)
+			pixmap_valid = (zoom_region_create_pixmap (zoom_region) == ZOOM_REGION_ERROR_NONE);
+		if (pixmap_valid)
+			zoom_region_update (zoom_region,
+					    zoom_region_source_rect_from_view_bounds (
+						    zoom_region,
+						    &zoom_region->priv->exposed_viewport));
+	}
 }
 
 static GdkRectangle
@@ -1060,7 +1084,7 @@ zoom_region_coalesce_updates (ZoomRegion *zoom_region)
 	int lookahead_n = 4; /* 'distance' to look ahead in queue */
 	int max_qlen = 50;
 
-	if (zoom_region->priv && g_list_length (zoom_region->priv->q) > max_qlen)
+	if (zoom_region->priv && zoom_region->priv->q && g_list_length (zoom_region->priv->q) > max_qlen)
 	{
 		g_list_free (zoom_region->priv->q);
 		zoom_region->priv->q = NULL; /* just discard and update everything */
@@ -1068,7 +1092,10 @@ zoom_region_coalesce_updates (ZoomRegion *zoom_region)
 		zoom_region_queue_update (zoom_region, zoom_region_rect_from_bounds
 			(zoom_region, &zoom_region->priv->source_area));
 	}
-	else if (zoom_region->priv && (g_list_length (zoom_region->priv->q) > 1) && can_coalesce)
+	else 
+
+        if (zoom_region->priv && zoom_region->priv->q && 
+	    (g_list_length (zoom_region->priv->q) > 1) && can_coalesce)
 	{		
 		q = g_list_reverse (g_list_copy (zoom_region->priv->q));
 		if (q)
@@ -1204,10 +1231,20 @@ zoom_region_create_pixmap (ZoomRegion *zoom_region)
 			   zoom_region->priv->exposed_viewport.x2,
 			   zoom_region->priv->exposed_viewport.y2);
 #endif
-		zoom_region_update (zoom_region,
-				    zoom_region_source_rect_from_view_bounds (
+
+			    DEBUG_RECT("viewport", zoom_region_source_rect_from_view_bounds
+					    (zoom_region, &zoom_region->priv->exposed_viewport));
+			    DEBUG_RECT("source", zoom_region_rect_from_bounds
+					    (zoom_region, &((Magnifier*)zoom_region->priv->parent)->source_bounds));
+
+			    zoom_region_update (zoom_region,
+/*				    zoom_region_source_rect_from_view_bounds (
 					    zoom_region,
 					    &zoom_region->priv->exposed_viewport));
+*/
+				    zoom_region_rect_from_bounds 
+				    (zoom_region, 
+				     &((Magnifier *)zoom_region->priv->parent)->source_bounds));
 		return ZOOM_REGION_ERROR_NONE;
 	}
 
@@ -1710,6 +1747,8 @@ zoom_region_get_source_subwindow (ZoomRegion *zoom_region,
 						 0,
 						 bounds.width,
 						 bounds.height);
+
+	/* TODO: blank the region overlapped by the target display if source == target */
 	
 	if (!subimage)
 		_debug_announce_rect ("update of invalid subregion!\n", bounds);
@@ -1763,12 +1802,17 @@ zoom_region_update (ZoomRegion *zoom_region,
 	{
 		GdkRectangle paint_rect;
 		gettimeofday(&mag_timing.scale_start, NULL);
+		DEBUG_RECT ("source rect", source_rect);
 		paint_rect = zoom_region_view_rect_from_source_rect (zoom_region, source_rect);
-		paint_rect = zoom_region_clip_to_window (zoom_region, paint_rect);
+		/* paint_rect = zoom_region_clip_to_scaled_pixmap (zoom_region, paint_rect); */
 		DEBUG_RECT ("paint rect", paint_rect);
 		if (zoom_region->invert)
 			_zoom_region_invert_pixbuf (subimage);
 
+		/** 
+		 *   XXX: We seem to be breaking with the original intention here, which was to 
+		 *   keep a fullscreen scaled pixbuf in-sync.
+		 **/
 		gdk_pixbuf_scale (subimage,
 				  zoom_region->priv->scaled_pixbuf,
 				  0,
@@ -1924,7 +1968,11 @@ zoom_region_process_updates (gpointer data)
 		return TRUE;
 	}
 	else 
+	{
+		if (zoom_region->priv) 
+			zoom_region->priv->update_handler_id = 0;
 		return FALSE;
+	}
 }
 
 void
@@ -2596,6 +2644,7 @@ impl_zoom_region_set_roi (PortableServer_Servant servant,
 		}
 
 		/* Set idle handler to process this pending update when possible */
+
 		pending_idle_handler = g_idle_add_full (GDK_PRIORITY_REDRAW + 2,
 			zoom_region_process_pending, zoom_region, NULL);
 
@@ -2648,10 +2697,10 @@ impl_zoom_region_set_mag_factor (PortableServer_Servant servant,
 
 	zoom_region->priv->scaled_pixbuf = gdk_pixbuf_new (
 		GDK_COLORSPACE_RGB, FALSE, 8,
-		zoom_region->priv->source_area.x2 -
-		zoom_region->priv->source_area.x1 * zoom_region->xscale + 1,
-		zoom_region->priv->source_area.y2 -
-		zoom_region->priv->source_area.y1 * zoom_region->yscale + 1);
+		(zoom_region->priv->source_area.x2 -
+		zoom_region->priv->source_area.x1) * zoom_region->xscale + 1,
+		(zoom_region->priv->source_area.y2 -
+		zoom_region->priv->source_area.y1) * zoom_region->yscale + 1);
 
 	if (zoom_region->priv->pixmap) {
 		g_object_unref (zoom_region->priv->pixmap);
@@ -2663,12 +2712,13 @@ impl_zoom_region_set_mag_factor (PortableServer_Servant servant,
 		zoom_region_create_pixmap (zoom_region);
 		g_object_unref (zoom_region->priv->scaled_pixbuf);
 
+		/* only create a scaled image big enough for the target display, for now */
 		zoom_region->priv->scaled_pixbuf = gdk_pixbuf_new (
 			GDK_COLORSPACE_RGB, FALSE, 8,
-			zoom_region->priv->source_area.x2 -
-			zoom_region->priv->source_area.x1 * zoom_region->xscale + 1,
-			zoom_region->priv->source_area.y2 -
-			zoom_region->priv->source_area.y1 * zoom_region->yscale + 1);
+			(zoom_region->priv->source_area.x2 -
+			zoom_region->priv->source_area.x1) * zoom_region->xscale + 1,
+			(zoom_region->priv->source_area.y2 -
+			zoom_region->priv->source_area.y1) * zoom_region->yscale + 1);
 		retval = CORBA_FALSE;
 	}
 	zoom_region_update_current (zoom_region);
@@ -2723,8 +2773,9 @@ impl_zoom_region_mark_dirty (PortableServer_Servant servant,
 			    zoom_region, roi_dirty) );
 
 	zoom_region_update_pointer (zoom_region, TRUE);
+	/* XXX ? should we clip here, or wait till process_updates? */
 	zoom_region_queue_update (zoom_region, 
-	  zoom_region_clip_to_target (zoom_region, 
+	  zoom_region_clip_to_source (zoom_region, 
 	      zoom_region_rect_from_bounds (zoom_region, roi_dirty)));
 }
 
@@ -2769,7 +2820,10 @@ zoom_region_do_dispose (ZoomRegion *zoom_region)
 	}
 	if (zoom_region->priv && zoom_region->priv->update_pointer_id)
 	    g_source_remove (zoom_region->priv->update_pointer_id);
+	if (zoom_region->priv && zoom_region->priv->update_handler_id)
+	    g_source_remove (zoom_region->priv->update_handler_id);
 	g_idle_remove_by_data (zoom_region);
+	
 #ifdef ZOOM_REGION_DEBUG
 	zoom_region->alive = FALSE;
 #endif
@@ -3082,6 +3136,7 @@ zoom_region_private_init (ZoomRegionPrivate *priv)
 	priv->exposed_viewport = rectbounds;
 	priv->source_area = rectbounds;
 	priv->update_pointer_id = 0;
+	priv->update_handler_id = 0;
 }
 
 static void
